%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TEMPORAL DEBLURRING SIMULATION
%
% This simulation program show a simple temporal deblurring example of a
% video of the globally shifted UC seal image at a variety of frame rates.
% This simulation program consists of four steps:
%
%   1. LOAD A TEST IMAGE
%      We use a pepper image as a test image in this simulation.
%
%   2. CREATE A SIMULATED GROUND-TRUTH SEQUENCE
%      We create a simulated ground-truth sequence by shifting the UC seal
%      image.
%
%   3. BLUR THE SIMULATED SEQUENCE WITH TEMPORAL PSF AND ADD SOME NOISE
%      The temporal PSF in this example is a 1x1x8 uniform PSF. Using the
%      PSF, we blur the simulated sequence and add weak noise (white
%      Gaussian noise with standard deviation 2). Assuming that the blurred
%      noisy sequence is the output of the space-time interpolator, the
%      next step is deblurring.
%
%   4. DEBLUR THE BLURRED NOISY SEQUENCE
%      The final step is deblurring. We use the true PSF in this example.
%      Using the steepest descent method, we iteratively estimate the
%      underlying sequence with 3-D (digital) total valiation.
%
% Note : An appropriate mirroring around the edges of the 3-D array of the
% videos is also necessary to avoid ringing. To simplify the simulation
% code, the mirroring is omitted from this code.
%
% wrote by Hiroyuki Takeda, Feb 9th, 2011


%%%%%%%%%%%%%%%%%%%%%%%%
% 1. LOAD A TEST IMAGE %
%%%%%%%%%%%%%%%%%%%%%%%%
UCseal = double(imread('Logos3666.bmp'));
img = UCseal(1:2:end, 1:2:end) + UCseal(2:2:end, 1:2:end) ...
    + UCseal(1:2:end, 2:2:end) + UCseal(2:2:end, 2:2:end);
img = img ./ max(img(:)) * 255;
img = img(35:300, 21:300);
[N M] = size(img);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2. CREATE A GROUND-TRUTH SEQUENCE %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
th = pi/180*6 : pi/180*6 : 6*pi;
radius = 10;
img = [ones(N, radius*2)*255, img, ones(N, radius*2)*255];
% generate motion shift vectors
motion = [zeros(1, length(th)); ...
          round(radius * sin(th))];
seq_true = zeros(N, M, length(th));
for i = 1 : length(th)
    % global shift (pure translational motion)
    seq_true(:,:,i) = img(1+motion(1,i):end+motion(1,i),...
                          1+radius*2+motion(2,i):end-radius*2+motion(2,i));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 3. BLUR THE SEQUENCE WITH A TEMPORAL PSF %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% temporal PSF (1x1x8 uniform PSF)
G = ones(1, 1, 9);
G(:,:,9) = 0;
G = G / sum(G(:));

% convolve the shifthing sequence with the temporal PSF 
seq_blur = convn(seq_true, G, 'same');
seq_true = seq_true(:,:,5:end-4);
seq_blur = seq_blur(:,:,5:end-4);

% add some noise
sigma = 0;         % no noise in this example
randn('state', 0);
seq_blur = seq_blur + randn(size(seq_blur)) * sigma;

% generate the blurred sequence in different frame rate
seq_blur_1fps = seq_blur(:,:,1:8:end); % the simulated input low-speed video     (1fsp)
seq_blur_2fps = seq_blur(:,:,1:4:end); % the simulated temporally upscaled video (2fps)
seq_blur_4fps = seq_blur(:,:,1:2:end); % the simulated temporally upscaled video (4fps)  
seq_blur_8fps = seq_blur(:,:,1:1:end); % the simulated temporally upscaled video (8fps)

% generate the corresponding PSF for each sequence
G_2fps = reshape([1/2, 1/2, 0], 1, 1, 3);                               % 1x1x2 uniform PSF
G_4fps = reshape([1/4, 1/4, 1/4, 1/4, 0], 1, 1, 5);                     % 1x1x4 uniform PSF
G_8fps = reshape([1/8, 1/8, 1/8, 1/8, 1/8, 1/8, 1/8, 1/8, 0], 1, 1, 9); % 1x1x8 uniform PSF

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% 4. DEBLURRING IN THE SPACE-TIME DOMAIN WITH THE TRUE 3-D PSF AND TOTAL VARIATION %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IT = 400;           % the number of iterations of the steepest descent
step = 1;         % the step size of the steepest descent update
lambda = 0.12;      % the regularization parameter
Gt_2fps = flipdim(G_2fps, 3); % the flipped version of the temporal PSF
Gt_4fps = flipdim(G_4fps, 3);
Gt_8fps = flipdim(G_8fps, 3);
u2fps = seq_blur_2fps;       % initialize with the blurred sequence
u4fps = seq_blur_4fps;
u8fps = seq_blur_8fps;

for it = 1 : IT
   
    % the gradient from the likelihood term
    ec2fps = convn(seq_blur_2fps - convn(u2fps, G_2fps, 'same'), Gt_2fps, 'same'); % G^T (y - G x)
    ec4fps = convn(seq_blur_4fps - convn(u4fps, G_4fps, 'same'), Gt_4fps, 'same');
    ec8fps = convn(seq_blur_8fps - convn(u8fps, G_8fps, 'same'), Gt_8fps, 'same');
    
    % the gradient from the TV regularization term
    % x1-direction
    tmp = sign(u2fps - ishift3(u2fps, 1, 0, 0)); tv2fps = tmp - ishift3(tmp, -1, 0, 0);
    tmp = sign(u4fps - ishift3(u4fps, 1, 0, 0)); tv4fps = tmp - ishift3(tmp, -1, 0, 0);
    tmp = sign(u8fps - ishift3(u8fps, 1, 0, 0)); tv8fps = tmp - ishift3(tmp, -1, 0, 0);
    % x2-direction
    tmp = sign(u2fps - ishift3(u2fps, 0, 1, 0)); tv2fps = tv2fps + tmp - ishift3(tmp, 0, -1, 0);
    tmp = sign(u4fps - ishift3(u4fps, 0, 1, 0)); tv4fps = tv4fps + tmp - ishift3(tmp, 0, -1, 0);
    tmp = sign(u8fps - ishift3(u8fps, 0, 1, 0)); tv8fps = tv8fps + tmp - ishift3(tmp, 0, -1, 0);
    % t-direction
    tmp = sign(u2fps - ishift3(u2fps, 0, 0, 1)); tv2fps = tv2fps + tmp - ishift3(tmp, 0, 0, -1);
    tmp = sign(u4fps - ishift3(u4fps, 0, 0, 1)); tv4fps = tv4fps + tmp - ishift3(tmp, 0, 0, -1);
    tmp = sign(u8fps - ishift3(u8fps, 0, 0, 1)); tv8fps = tv8fps + tmp - ishift3(tmp, 0, 0, -1);
    
    % update
    update2fps = step * (ec2fps - (lambda * 4) * tv2fps); update2fps(:,:,1:3) = 0; update2fps(:,:,end-2:end) = 0;
    update4fps = step * (ec4fps - (lambda * 2) * tv4fps); update4fps(:,:,1:5) = 0; update4fps(:,:,end-4:end) = 0;
    update8fps = step * (ec8fps - lambda * tv8fps); update8fps(:,:,1:9) = 0; update8fps(:,:,end-8:end) = 0;
    u2fps = u2fps + update2fps;
    u4fps = u4fps + update4fps;
    u8fps = u8fps + update8fps;
    
    % display the estimation status
    fnum = 32;
    figure(1); 
    subplot(3,1,1); imagesc(uint8([u2fps(:,:,fnum/4), u4fps(:,:,fnum/2), u8fps(:,:,fnum)])); colormap(gray); axis image; colorbar;
    title(sprintf('The current estimate of the frames of 2fps (left), 4fps (center), 8fps (right) videos at t=%d, %d iterations', fnum, it));
    subplot(3,1,2); imagesc([ec2fps(:,:,fnum/4), ec4fps(:,:,fnum/2), ec8fps(:,:,fnum)]); colormap(gray); axis image; colorbar;
    title('The gradient of the likelihood term');
    subplot(3,1,3); imagesc([tv2fps(:,:,fnum/4), tv4fps(:,:,fnum/2), tv8fps(:,:,fnum)]); colormap(gray); axis image; colorbar;
    title('The gradient of the total valiation term');
    pause(0.1);
end

% display the results
for j = 1 : 10
    for i = 41 : 100

        figure(3);
        subplot(3,1,1); imagesc(uint8([seq_true(:,:,i), seq_blur_1fps(:,:,ceil(i/8)), ones(size(seq_true(:,:,1))) * 255]));
        colormap(gray); axis image; axis off;
        title('(left) The original video at 8fps, (right) A blurry low speed video at 1fps.');
        subplot(3,1,2); imagesc(uint8([seq_blur_2fps(:,:,ceil(i/4)), seq_blur_4fps(:,:,ceil(i/2)), seq_blur_8fps(:,:,i)]));
        colormap(gray); axis image; axis off;
        title('Temporally upscaled videos of the low speed (1fps) video at (left) 2fps, (center) 4fps, and (right) 8fps.');
        subplot(3,1,3); imagesc(uint8([u2fps(:,:,ceil(i/4)), u4fps(:,:,ceil(i/2)), u8fps(:,:,i)]));
        colormap(gray); axis image; axis off;
        title('Deblurred videos by our 3-D deblurrng methods at (left) 2fps, (center) 4fps, and (right) 8fps.');
        pause(0.1);
    end
end



